==> neuroflex-0.1.3.tar.gz: neuroflex-0.1.3/NeuroFlex/advanced_models/advanced_math_solving.py <==
import numpy as np
from scipy import optimize, integrate
from typing import List, Dict, Any, Union
import time
import logging
import torch
import torch.nn as nn
import torch.optim as optim
from ..constants import PERFORMANCE_THRESHOLD, UPDATE_INTERVAL, LEARNING_RATE_ADJUSTMENT, MAX_HEALING_ATTEMPTS

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class AdvancedMathSolver:
    def __init__(self):
        self.available_methods = {
            'linear_algebra': self.solve_linear_algebra,
            'calculus': self.solve_calculus,
            'optimization': self.solve_optimization,
            'differential_equations': self.solve_differential_equations
        }
        self.performance = 0.0
        self.last_update = time.time()
        self.performance_history = []
        self.learning_rate = 0.001
        self.performance_threshold = PERFORMANCE_THRESHOLD
        self.update_interval = UPDATE_INTERVAL
        self.max_healing_attempts = MAX_HEALING_ATTEMPTS
        self.rl_agent = RLAgent(state_dim=3, action_dim=1)  # Initialize RL agent

    def solve(self, problem_type: str, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Solve an advanced math problem based on the problem type and data.

        Args:
            problem_type (str): Type of math problem (e.g., 'linear_algebra', 'calculus')
            problem_data (Dict[str, Any]): Dictionary containing problem-specific data

        Returns:
            Dict[str, Any]: Solution to the problem
        """
        if problem_type not in self.available_methods:
            raise ValueError(f"Unsupported problem type: {problem_type}")

        try:
            solution = self.available_methods[problem_type](problem_data)
            self._update_performance(1.0)  # Assume perfect performance for successful solve
            return solution
        except Exception as e:
            self._update_performance(0.0)  # Performance is 0 if an error occurs
            logger.error(f"Error solving {problem_type} problem: {str(e)}")
            raise
        finally:
            if self.performance < self.performance_threshold:
                self._self_heal()

    def solve_linear_algebra(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve linear algebra problems."""
        if 'matrix_a' in problem_data and 'vector_b' in problem_data:
            A = problem_data['matrix_a']
            b = problem_data['vector_b']
            x = np.linalg.solve(A, b)
            return {'solution': x}
        elif 'matrix' in problem_data:
            A = problem_data['matrix']
            eigenvalues, eigenvectors = np.linalg.eig(A)
            return {'eigenvalues': eigenvalues, 'eigenvectors': eigenvectors}
        else:
            raise ValueError("Unsupported linear algebra problem")

    def solve_calculus(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve calculus problems."""
        if 'function' in problem_data and 'variable' in problem_data:
            def f(x):
                return eval(problem_data['function'])

            a, b = problem_data.get('interval', (-np.inf, np.inf))
            integral, error = integrate.quad(f, a, b)
            return {'integral': integral, 'error': error}
        else:
            raise ValueError("Unsupported calculus problem")

    def solve_optimization(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve optimization problems."""
        if 'function' in problem_data and 'initial_guess' in problem_data:
            def f(x):
                return eval(problem_data['function'])

            x0 = problem_data['initial_guess']
            result = optimize.minimize(f, x0)
            return {'optimal_x': result.x, 'optimal_value': result.fun}
        else:
            raise ValueError("Unsupported optimization problem")

    def solve_differential_equations(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve differential equations."""
        if 'function' in problem_data and 'initial_conditions' in problem_data:
            def f(t, y):
                return eval(problem_data['function'])

            t_span = problem_data.get('t_span', (0, 10))
            y0 = problem_data['initial_conditions']

            # Use RK45 method with increased number of time points
            t_eval = np.linspace(t_span[0], t_span[1], 1000)
            solution = integrate.solve_ivp(f, t_span, y0, method='RK45', t_eval=t_eval, rtol=1e-8, atol=1e-8)
            return {'t': solution.t, 'y': solution.y}
        else:
            raise ValueError("Unsupported differential equation problem")

    def _update_performance(self, new_performance: float):
        """Update the performance history and trigger self-healing if necessary."""
        self.performance = new_performance
        self.performance_history.append(new_performance)
        if len(self.performance_history) > 100:
            self.performance_history.pop(0)
        self.last_update = time.time()

    def _self_heal(self):
        """Implement self-healing mechanisms using RL agent."""
        logger.info("Initiating self-healing process...")
        initial_performance = self.performance
        best_performance = initial_performance
        select_action_calls = 0
        update_calls = 0

        for attempt in range(self.max_healing_attempts):
            state = self._get_state()
            action = self.rl_agent.select_action(state)
            select_action_calls += 1
            self._apply_action(action)
            new_performance = self._simulate_performance()

            if new_performance > best_performance:
                best_performance = new_performance
                self.performance = best_performance
                logger.info(f"Performance improved: {self.performance:.4f}")

            reward = new_performance - initial_performance
            next_state = self._get_state()
            self.rl_agent.update(state, action, reward, next_state, new_performance >= self.performance_threshold)
            update_calls += 1

            if new_performance >= self.performance_threshold:
                self.performance = new_performance
                logger.info(f"Self-healing successful after {attempt + 1} attempts.")
                logger.info(f"RL agent method calls - select_action: {select_action_calls}, update: {update_calls}")
                return

        if best_performance > initial_performance:
            self.performance = best_performance
            logger.info(f"Self-healing improved performance. New performance: {self.performance:.4f}")
        else:
            logger.warning("Self-healing not improving performance. Performance unchanged.")
            self.performance = initial_performance

        logger.info(f"Self-healing completed. Final performance: {self.performance:.4f}")
        logger.info(f"RL agent method calls - select_action: {select_action_calls}, update: {update_calls}")

    def _get_state(self) -> np.ndarray:
        """Get the current state for the RL agent."""
        return np.array([self.performance, self.learning_rate, len(self.performance_history)])

    def _apply_action(self, action: float):
        """Apply the action suggested by the RL agent."""
        self.learning_rate = max(min(self.learning_rate * (1 + action), 0.1), 1e-5)
        logger.info(f"Adjusted learning rate to {self.learning_rate:.6f}")

    def diagnose(self) -> List[str]:
        """Diagnose potential issues with the solver."""
        issues = []
        if self.performance < self.performance_threshold:
            issues.append(f"Low performance: {self.performance:.4f}")
        time_since_update = time.time() - self.last_update
        if time_since_update > self.update_interval:
            issues.append(f"Long time since last update: {time_since_update / 3600:.2f} hours")
        if len(self.performance_history) >= 5 and all(p < self.performance_threshold for p in self.performance_history[-5:]):
            issues.append("Consistently low performance")
        return issues

    def _simulate_performance(self) -> float:
        """Simulate new performance after applying healing strategies."""
        # This is a placeholder. In a real scenario, you would re-evaluate the solver's performance.
        return self.performance * (1 + np.random.uniform(-0.1, 0.1))

class RLAgent:
    def __init__(self, state_dim: int, action_dim: int):
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.policy_net = nn.Sequential(
            nn.Linear(state_dim, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, action_dim),
            nn.Tanh()
        )
        self.optimizer = optim.Adam(self.policy_net.parameters(), lr=0.001)

    def select_action(self, state: np.ndarray) -> float:
        state_tensor = torch.FloatTensor(state)
        with torch.no_grad():
            action = self.policy_net(state_tensor).item()
        return action

    def update(self, state: np.ndarray, action: float, reward: float, next_state: np.ndarray, done: bool):
        state_tensor = torch.FloatTensor(state)
        action_tensor = torch.FloatTensor([action])
        reward_tensor = torch.FloatTensor([reward])
        next_state_tensor = torch.FloatTensor(next_state)

        # Compute the loss
        predicted_action = self.policy_net(state_tensor)
        loss = nn.MSELoss()(predicted_action, action_tensor)

        # Backpropagate and update the policy
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

# Example usage
if __name__ == "__main__":
    solver = AdvancedMathSolver()

    # Example linear algebra problem
    linear_algebra_problem = {
        'matrix_a': np.array([[1, 2], [3, 4]]),
        'vector_b': np.array([5, 6])
    }

    solution = solver.solve('linear_algebra', linear_algebra_problem)
    print("Linear Algebra Solution:", solution)

    # Demonstrate self-healing
    solver.performance = 0.5  # Set a low performance to trigger self-healing
    issues = solver.diagnose()
    print("Diagnosed issues:", issues)
    solver._self_heal()
    print("Performance after self-healing:", solver.performance)

==> neuroflex-0.1.4.tar.gz: neuroflex-0.1.4/NeuroFlex/advanced_models/advanced_math_solving.py <==
# MIT License
# 
# Copyright (c) 2024 VishwamAI
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from scipy import optimize, integrate
from typing import List, Dict, Any, Union
import time
import logging
import torch
import torch.nn as nn
import torch.optim as optim
from ..constants import PERFORMANCE_THRESHOLD, UPDATE_INTERVAL, LEARNING_RATE_ADJUSTMENT, MAX_HEALING_ATTEMPTS

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class AdvancedMathSolver:
    def __init__(self):
        self.available_methods = {
            'linear_algebra': self.solve_linear_algebra,
            'calculus': self.solve_calculus,
            'optimization': self.solve_optimization,
            'differential_equations': self.solve_differential_equations
        }
        self.performance = 0.0
        self.last_update = time.time()
        self.performance_history = []
        self.learning_rate = 0.001
        self.performance_threshold = PERFORMANCE_THRESHOLD
        self.update_interval = UPDATE_INTERVAL
        self.max_healing_attempts = MAX_HEALING_ATTEMPTS
        self.rl_agent = RLAgent(state_dim=3, action_dim=1)  # Initialize RL agent

    def solve(self, problem_type: str, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Solve an advanced math problem based on the problem type and data.

        Args:
            problem_type (str): Type of math problem (e.g., 'linear_algebra', 'calculus')
            problem_data (Dict[str, Any]): Dictionary containing problem-specific data

        Returns:
            Dict[str, Any]: Solution to the problem
        """
        if problem_type not in self.available_methods:
            raise ValueError(f"Unsupported problem type: {problem_type}")

        try:
            solution = self.available_methods[problem_type](problem_data)
            self._update_performance(1.0)  # Assume perfect performance for successful solve
            return solution
        except Exception as e:
            self._update_performance(0.0)  # Performance is 0 if an error occurs
            logger.error(f"Error solving {problem_type} problem: {str(e)}")
            raise
        finally:
            if self.performance < self.performance_threshold:
                self._self_heal()

    def solve_linear_algebra(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve linear algebra problems."""
        if 'matrix_a' in problem_data and 'vector_b' in problem_data:
            A = problem_data['matrix_a']
            b = problem_data['vector_b']
            x = np.linalg.solve(A, b)
            return {'solution': x}
        elif 'matrix' in problem_data:
            A = problem_data['matrix']
            eigenvalues, eigenvectors = np.linalg.eig(A)
            return {'eigenvalues': eigenvalues, 'eigenvectors': eigenvectors}
        else:
            raise ValueError("Unsupported linear algebra problem")

    def solve_calculus(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve calculus problems."""
        if 'function' in problem_data and 'variable' in problem_data:
            def f(x):
                return eval(problem_data['function'])

            a, b = problem_data.get('interval', (-np.inf, np.inf))
            integral, error = integrate.quad(f, a, b)
            return {'integral': integral, 'error': error}
        else:
            raise ValueError("Unsupported calculus problem")

    def solve_optimization(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve optimization problems."""
        if 'function' in problem_data and 'initial_guess' in problem_data:
            def f(x):
                # Use a safer alternative to eval
                return problem_data['function'](x)

            x0 = problem_data['initial_guess']
            # Use BFGS method with increased max iterations
            result = optimize.minimize(f, x0, method='BFGS', options={'maxiter': 1000})
            return {'optimal_x': result.x, 'optimal_value': result.fun, 'success': result.success, 'message': result.message}
        else:
            raise ValueError("Unsupported optimization problem")

    def solve_differential_equations(self, problem_data: Dict[str, Any]) -> Dict[str, Any]:
        """Solve differential equations."""
        if 'function' in problem_data and 'initial_conditions' in problem_data:
            def f(t, y):
                return eval(problem_data['function'])

            t_span = problem_data.get('t_span', (0, 10))
            y0 = problem_data['initial_conditions']

            # Use RK45 method with increased number of time points
            t_eval = np.linspace(t_span[0], t_span[1], 1000)
            solution = integrate.solve_ivp(f, t_span, y0, method='RK45', t_eval=t_eval, rtol=1e-8, atol=1e-8)
            return {'t': solution.t, 'y': solution.y}
        else:
            raise ValueError("Unsupported differential equation problem")

    def _update_performance(self, new_performance: float):
        """Update the performance history and trigger self-healing if necessary."""
        self.performance = new_performance
        self.performance_history.append(new_performance)
        if len(self.performance_history) > 100:
            self.performance_history.pop(0)
        self.last_update = time.time()

    def _self_heal(self):
        """Implement self-healing mechanisms using RL agent."""
        logger.info("Initiating self-healing process...")
        initial_performance = self.performance
        best_performance = initial_performance
        select_action_calls = 0
        update_calls = 0

        for attempt in range(self.max_healing_attempts):
            state = self._get_state()
            action = self.rl_agent.select_action(state)
            select_action_calls += 1
            self._apply_action(action)
            new_performance = self._simulate_performance()

            if new_performance > best_performance:
                best_performance = new_performance
                self.performance = best_performance
                logger.info(f"Performance improved: {self.performance:.4f}")

            reward = new_performance - initial_performance
            next_state = self._get_state()
            self.rl_agent.update(state, action, reward, next_state, new_performance >= self.performance_threshold)
            update_calls += 1

            if new_performance >= self.performance_threshold:
                self.performance = new_performance
                logger.info(f"Self-healing successful after {attempt + 1} attempts.")
                logger.info(f"RL agent method calls - select_action: {select_action_calls}, update: {update_calls}")
                return

        if best_performance > initial_performance:
            self.performance = best_performance
            logger.info(f"Self-healing improved performance. New performance: {self.performance:.4f}")
        else:
            logger.warning("Self-healing not improving performance. Performance unchanged.")
            self.performance = initial_performance

        logger.info(f"Self-healing completed. Final performance: {self.performance:.4f}")
        logger.info(f"RL agent method calls - select_action: {select_action_calls}, update: {update_calls}")

    def _get_state(self) -> np.ndarray:
        """Get the current state for the RL agent."""
        return np.array([self.performance, self.learning_rate, len(self.performance_history)])

    def _apply_action(self, action: float):
        """Apply the action suggested by the RL agent."""
        self.learning_rate = max(min(self.learning_rate * (1 + action), 0.1), 1e-5)
        logger.info(f"Adjusted learning rate to {self.learning_rate:.6f}")

    def diagnose(self) -> List[str]:
        """Diagnose potential issues with the solver."""
        issues = []
        if self.performance < self.performance_threshold:
            issues.append(f"Low performance: {self.performance:.4f}")
        time_since_update = time.time() - self.last_update
        if time_since_update > self.update_interval:
            issues.append(f"Long time since last update: {time_since_update / 3600:.2f} hours")
        if len(self.performance_history) >= 5 and all(p < self.performance_threshold for p in self.performance_history[-5:]):
            issues.append("Consistently low performance")
        return issues

    def _simulate_performance(self) -> float:
        """Simulate new performance after applying healing strategies."""
        # This is a placeholder. In a real scenario, you would re-evaluate the solver's performance.
        return self.performance * (1 + np.random.uniform(-0.1, 0.1))

class RLAgent:
    def __init__(self, state_dim: int, action_dim: int):
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.policy_net = nn.Sequential(
            nn.Linear(state_dim, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, action_dim),
            nn.Tanh()
        )
        self.optimizer = optim.Adam(self.policy_net.parameters(), lr=0.001)

    def select_action(self, state: np.ndarray) -> float:
        state_tensor = torch.FloatTensor(state)
        with torch.no_grad():
            action = self.policy_net(state_tensor).item()
        return action

    def update(self, state: np.ndarray, action: float, reward: float, next_state: np.ndarray, done: bool):
        state_tensor = torch.FloatTensor(state)
        action_tensor = torch.FloatTensor([action])
        reward_tensor = torch.FloatTensor([reward])
        next_state_tensor = torch.FloatTensor(next_state)

        # Compute the loss
        predicted_action = self.policy_net(state_tensor)
        loss = nn.MSELoss()(predicted_action, action_tensor)

        # Backpropagate and update the policy
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

# Example usage
if __name__ == "__main__":
    solver = AdvancedMathSolver()

    # Example linear algebra problem
    linear_algebra_problem = {
        'matrix_a': np.array([[1, 2], [3, 4]]),
        'vector_b': np.array([5, 6])
    }

    solution = solver.solve('linear_algebra', linear_algebra_problem)
    print("Linear Algebra Solution:", solution)

    # Demonstrate self-healing
    solver.performance = 0.5  # Set a low performance to trigger self-healing
    issues = solver.diagnose()
    print("Diagnosed issues:", issues)
    solver._self_heal()
    print("Performance after self-healing:", solver.performance)
